Package: SeqArray
Type: Package
Title: Data management of large-scale whole-genome sequence variant
        calls using GDS files
Version: 1.52.1
Date: 2026-05-20
Depends: R (>= 3.5.0), gdsfmt (>= 1.31.1)
Imports: methods, parallel, digest, S4Vectors, IRanges, GenomicRanges,
        Seqinfo, Biostrings
LinkingTo: gdsfmt
Suggests: Biobase, BiocGenerics, BiocParallel, RUnit, Rcpp, SNPRelate,
        crayon, knitr, markdown, rmarkdown, Rsamtools,
        VariantAnnotation
Authors@R: c(person("Xiuwen", "Zheng", role=c("aut", "cre"),
        email="zhengx@u.washington.edu", comment=c(ORCID="0000-0002-1390-0708")),
        person("Stephanie", "Gogarten", role="aut", email="sdmorris@uw.edu"),
        person("David", "Levine", role="ctb"),
        person("Cathy", "Laurie", role="ctb"))
Description: Data management of large-scale whole-genome sequencing variant
        calls with thousands of individuals: genotypic data (e.g., SNVs, indels
        and structural variation calls) and annotations in SeqArray GDS files
        are stored in an array-oriented and compressed manner, with efficient
        data access using the R programming language.
License: GPL-3
VignetteBuilder: knitr
ByteCompile: TRUE
LazyData: true
URL: https://github.com/zhengxwen/SeqArray
BugReports: https://github.com/zhengxwen/SeqArray/issues
biocViews: Infrastructure, DataRepresentation, Sequencing, Genetics
git_url: https://git.bioconductor.org/packages/SeqArray
git_branch: RELEASE_3_23
git_last_commit: fbd42ac
git_last_commit_date: 2026-05-30
Repository: Bioconductor 3.23
Date/Publication: 2026-05-31
NeedsCompilation: yes
Packaged: 2026-06-01 01:19:50 UTC; biocbuild
Author: Xiuwen Zheng [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-1390-0708>),
  Stephanie Gogarten [aut],
  David Levine [ctb],
  Cathy Laurie [ctb]
Maintainer: Xiuwen Zheng <zhengx@u.washington.edu>
